##################################################
# ./AppService_services.py
# generated by ZSI.wsdl2python
#
#
##################################################


from AppService_services_types import *
from AppService_services_types import nbcr_sdsc_edu_opal_types as ns1
import urlparse, types
from ZSI.TCcompound import Struct
from ZSI import client
import ZSI


class AppServiceInterface:
    def getAppServicePortType(self, portAddress=None, **kw):
        raise NonImplementationError, "method not implemented"


class AppServiceLocator(AppServiceInterface):
    AppServicePortType_address = (
        "http://ws.nbcr.net:8080/axis-1.2.1/services/Pdb2pqrOpalService"
    )

    def getAppServicePortTypeAddress(self):
        return AppServiceLocator.AppServicePortType_address

    def getAppServicePortType(self, portAddress=None, **kw):
        return AppServicePortSoapBindingSOAP(
            portAddress or AppServiceLocator.AppServicePortType_address, **kw
        )


class AppServicePortSoapBindingSOAP:
    def __init__(self, addr, **kw):
        netloc = (urlparse.urlparse(addr)[1]).split(":") + [
            80,
        ]
        if not kw.has_key("host"):
            kw["host"] = netloc[0]
        if not kw.has_key("port"):
            kw["port"] = int(netloc[1])
        if not kw.has_key("url"):
            kw["url"] = urlparse.urlparse(addr)[2]
        self.binding = client.Binding(**kw)

    def destroy(self, request):
        """
        @param: request is str

        @return: response from destroyResponse::
            _destroyOutput: ns1.StatusOutputType_Def
              _baseURL: str
              _code: int
              _message: str
        """

        if not isinstance(request, basestring):
            raise TypeError, "%s incorrect request type" % (request.__class__)
        kw = {"requestclass": destroyRequestWrapper}
        response = self.binding.Send(
            None, None, request, soapaction="http://nbcr.sdsc.edu/opal/destroy", **kw
        )
        response = self.binding.Receive(destroyResponseWrapper())

        if not isinstance(response, destroyResponse) and not issubclass(
            destroyResponse, response.__class__
        ):
            raise TypeError, "%s incorrect response type" % (response.__class__)
        return response

    def getAppConfig(self, request):
        """
        @param: request to getAppConfigRequest::
            _getAppConfigInput: ns1.AppConfigInputType_Def

        @return: response from getAppConfigResponse::
            _getAppConfigOutput: ns1.AppConfigType_Def
              _binaryLocation: str
              _defaultArgs: str, optional
              _metadata: ns1.AppMetadataType_Def
                _info: str, optional
                _types: ns1.ArgumentsType_Def, optional
                  _flags: ns1.FlagsArrayType_Def, optional
                    _flag: ns1.FlagsType_Def, optional
                      _id: str
                      _tag: str
                      _textDesc: str, optional
                  _implicitParams: ns1.ImplicitParamsArrayType_Def, optional
                    _param: ns1.ImplicitParamsType_Def, optional
                      _extension: str, optional
                      _id: str
                      _ioType: ns1.IOType_Def
                        _IOType: str, optional
                      _max: int, optional
                      _min: int, optional
                      _name: str, optional
                      _required: boolean, optional
                      _semanticType: str, optional
                      _textDesc: str, optional
                  _taggedParams: ns1.ParamsArrayType_Def, optional
                    _param: ns1.ParamsType_Def, optional
                      _id: str
                      _ioType: ns1.IOType_Def, optional
                      _paramType: ns1.ParamType_Def
                        _ParamType: str, optional
                      _required: boolean, optional
                      _semanticType: str, optional
                      _tag: str, optional
                      _textDesc: str, optional
                      _value: str, optional
                    _separator: str, optional
                  _untaggedParams: ns1.ParamsArrayType_Def, optional
                _usage: str
              _parallel: boolean
        """

        if not isinstance(request, getAppConfigRequest) and not issubclass(
            getAppConfigRequest, request.__class__
        ):
            raise TypeError, "%s incorrect request type" % (request.__class__)
        kw = {}
        response = self.binding.Send(
            None,
            None,
            request,
            soapaction="http://nbcr.sdsc.edu/opal/getAppConfig",
            **kw
        )
        response = self.binding.Receive(getAppConfigResponseWrapper())

        if not isinstance(response, getAppConfigResponse) and not issubclass(
            getAppConfigResponse, response.__class__
        ):
            raise TypeError, "%s incorrect response type" % (response.__class__)
        return response

    def getAppMetadata(self, request):
        """
        @param: request to getAppMetadataRequest::
            _getAppMetadataInput: ns1.AppMetadataInputType_Def

        @return: response from getAppMetadataResponse::
            _getAppMetadataOutput: ns1.AppMetadataType_Def
              _info: str, optional
              _types: ns1.ArgumentsType_Def, optional
                _flags: ns1.FlagsArrayType_Def, optional
                  _flag: ns1.FlagsType_Def, optional
                    _id: str
                    _tag: str
                    _textDesc: str, optional
                _implicitParams: ns1.ImplicitParamsArrayType_Def, optional
                  _param: ns1.ImplicitParamsType_Def, optional
                    _extension: str, optional
                    _id: str
                    _ioType: ns1.IOType_Def
                      _IOType: str, optional
                    _max: int, optional
                    _min: int, optional
                    _name: str, optional
                    _required: boolean, optional
                    _semanticType: str, optional
                    _textDesc: str, optional
                _taggedParams: ns1.ParamsArrayType_Def, optional
                  _param: ns1.ParamsType_Def, optional
                    _id: str
                    _ioType: ns1.IOType_Def, optional
                    _paramType: ns1.ParamType_Def
                      _ParamType: str, optional
                    _required: boolean, optional
                    _semanticType: str, optional
                    _tag: str, optional
                    _textDesc: str, optional
                    _value: str, optional
                  _separator: str, optional
                _untaggedParams: ns1.ParamsArrayType_Def, optional
              _usage: str
        """

        if not isinstance(request, getAppMetadataRequest) and not issubclass(
            getAppMetadataRequest, request.__class__
        ):
            raise TypeError, "%s incorrect request type" % (request.__class__)
        kw = {}
        response = self.binding.Send(
            None,
            None,
            request,
            soapaction="http://nbcr.sdsc.edu/opal/getAppMetadata",
            **kw
        )
        response = self.binding.Receive(getAppMetadataResponseWrapper())

        if not isinstance(response, getAppMetadataResponse) and not issubclass(
            getAppMetadataResponse, response.__class__
        ):
            raise TypeError, "%s incorrect response type" % (response.__class__)
        return response

    def getOutputAsBase64ByName(self, request):
        """
        @param: request to getOutputAsBase64ByNameRequest::
            _getOutputAsBase64ByNameInput: ns1.OutputsByNameInputType_Def
              _fileName: str
              _jobID: str

        @return: response from getOutputAsBase64ByNameResponse::
            _item: str, optional
        """

        if not isinstance(request, getOutputAsBase64ByNameRequest) and not issubclass(
            getOutputAsBase64ByNameRequest, request.__class__
        ):
            raise TypeError, "%s incorrect request type" % (request.__class__)
        kw = {}
        response = self.binding.Send(
            None,
            None,
            request,
            soapaction="http://nbcr.sdsc.edu/opal/getOutputAsBase64ByName",
            **kw
        )
        response = self.binding.Receive(getOutputAsBase64ByNameResponseWrapper())

        if not isinstance(response, getOutputAsBase64ByNameResponse) and not issubclass(
            getOutputAsBase64ByNameResponse, response.__class__
        ):
            raise TypeError, "%s incorrect response type" % (response.__class__)
        return response

    def getOutputs(self, request):
        """
        @param: request is str

        @return: response from getOutputsResponse::
            _getOutputsOutput: ns1.JobOutputType_Def
              _outputFile: ns1.OutputFileType_Def, optional
                _name: str
                _url: str
              _stdErr: str, optional
              _stdOut: str, optional
        """

        if not isinstance(request, basestring):
            raise TypeError, "%s incorrect request type" % (request.__class__)
        kw = {"requestclass": getOutputsRequestWrapper}
        response = self.binding.Send(
            None, None, request, soapaction="http://nbcr.sdsc.edu/opal/getOutputs", **kw
        )
        response = self.binding.Receive(getOutputsResponseWrapper())

        if not isinstance(response, getOutputsResponse) and not issubclass(
            getOutputsResponse, response.__class__
        ):
            raise TypeError, "%s incorrect response type" % (response.__class__)
        return response

    def launchJob(self, request):
        """
        @param: request to launchJobRequest::
            _launchJobInput: ns1.JobInputType_Def
              _argList: str, optional
              _inputFile: ns1.InputFileType_Def, optional
                _contents: str
                _name: str
              _numProcs: int, optional

        @return: response from launchJobResponse::
            _launchJobOutput: ns1.JobSubOutputType_Def
              _jobID: str
              _status: ns1.StatusOutputType_Def
                _baseURL: str
                _code: int
                _message: str
        """

        if not isinstance(request, launchJobRequest) and not issubclass(
            launchJobRequest, request.__class__
        ):
            raise TypeError, "%s incorrect request type" % (request.__class__)
        kw = {}
        response = self.binding.Send(
            None, None, request, soapaction="http://nbcr.sdsc.edu/opal/launchJob", **kw
        )
        response = self.binding.Receive(launchJobResponseWrapper())

        if not isinstance(response, launchJobResponse) and not issubclass(
            launchJobResponse, response.__class__
        ):
            raise TypeError, "%s incorrect response type" % (response.__class__)
        return response

    def launchJobBlocking(self, request):
        """
        @param: request to launchJobBlockingRequest::
            _launchJobBlockingInput: ns1.JobInputType_Def
              _argList: str, optional
              _inputFile: ns1.InputFileType_Def, optional
                _contents: str
                _name: str
              _numProcs: int, optional

        @return: response from launchJobBlockingResponse::
            _launchJobBlockingOutput: ns1.BlockingOutputType_Def
              _jobOut: ns1.JobOutputType_Def
                _outputFile: ns1.OutputFileType_Def, optional
                  _name: str
                  _url: str
                _stdErr: str, optional
                _stdOut: str, optional
              _status: ns1.StatusOutputType_Def
                _baseURL: str
                _code: int
                _message: str
        """

        if not isinstance(request, launchJobBlockingRequest) and not issubclass(
            launchJobBlockingRequest, request.__class__
        ):
            raise TypeError, "%s incorrect request type" % (request.__class__)
        kw = {}
        response = self.binding.Send(
            None,
            None,
            request,
            soapaction="http://nbcr.sdsc.edu/opal/launchJobBlocking",
            **kw
        )
        response = self.binding.Receive(launchJobBlockingResponseWrapper())

        if not isinstance(response, launchJobBlockingResponse) and not issubclass(
            launchJobBlockingResponse, response.__class__
        ):
            raise TypeError, "%s incorrect response type" % (response.__class__)
        return response

    def queryStatus(self, request):
        """
        @param: request is str

        @return: response from queryStatusResponse::
            _queryStatusOutput: ns1.StatusOutputType_Def
              _baseURL: str
              _code: int
              _message: str
        """

        if not isinstance(request, basestring):
            raise TypeError, "%s incorrect request type" % (request.__class__)
        kw = {"requestclass": queryStatusRequestWrapper}
        response = self.binding.Send(
            None,
            None,
            request,
            soapaction="http://nbcr.sdsc.edu/opal/queryStatus",
            **kw
        )
        response = self.binding.Receive(queryStatusResponseWrapper())

        if not isinstance(response, queryStatusResponse) and not issubclass(
            queryStatusResponse, response.__class__
        ):
            raise TypeError, "%s incorrect response type" % (response.__class__)
        return response


class destroyRequest(ns1.destroyInput_Dec):
    if not hasattr(ns1.destroyInput_Dec(), "typecode"):
        typecode = ns1.destroyInput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.destroyInput_Dec.__init__(self, name=None, ns=None)


class destroyRequestWrapper(destroyRequest):
    """wrapper for document:literal message"""

    typecode = destroyRequest(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        destroyRequest.__init__(self, name=None, ns=None)


class destroyResponse(ns1.destroyOutput_Dec):
    if not hasattr(ns1.destroyOutput_Dec(), "typecode"):
        typecode = ns1.destroyOutput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.destroyOutput_Dec.__init__(self, name=None, ns=None)


class destroyResponseWrapper(destroyResponse):
    """wrapper for document:literal message"""

    typecode = destroyResponse(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        destroyResponse.__init__(self, name=None, ns=None)


class getAppConfigRequest(ns1.getAppConfigInput_Dec):
    if not hasattr(ns1.getAppConfigInput_Dec(), "typecode"):
        typecode = ns1.getAppConfigInput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.getAppConfigInput_Dec.__init__(self, name=None, ns=None)


class getAppConfigRequestWrapper(getAppConfigRequest):
    """wrapper for document:literal message"""

    typecode = getAppConfigRequest(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        getAppConfigRequest.__init__(self, name=None, ns=None)


class getAppConfigResponse(ns1.getAppConfigOutput_Dec):
    if not hasattr(ns1.getAppConfigOutput_Dec(), "typecode"):
        typecode = ns1.getAppConfigOutput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.getAppConfigOutput_Dec.__init__(self, name=None, ns=None)


class getAppConfigResponseWrapper(getAppConfigResponse):
    """wrapper for document:literal message"""

    typecode = getAppConfigResponse(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        getAppConfigResponse.__init__(self, name=None, ns=None)


class getAppMetadataRequest(ns1.getAppMetadataInput_Dec):
    if not hasattr(ns1.getAppMetadataInput_Dec(), "typecode"):
        typecode = ns1.getAppMetadataInput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.getAppMetadataInput_Dec.__init__(self, name=None, ns=None)


class getAppMetadataRequestWrapper(getAppMetadataRequest):
    """wrapper for document:literal message"""

    typecode = getAppMetadataRequest(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        getAppMetadataRequest.__init__(self, name=None, ns=None)


class getAppMetadataResponse(ns1.getAppMetadataOutput_Dec):
    if not hasattr(ns1.getAppMetadataOutput_Dec(), "typecode"):
        typecode = ns1.getAppMetadataOutput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.getAppMetadataOutput_Dec.__init__(self, name=None, ns=None)


class getAppMetadataResponseWrapper(getAppMetadataResponse):
    """wrapper for document:literal message"""

    typecode = getAppMetadataResponse(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        getAppMetadataResponse.__init__(self, name=None, ns=None)


class getOutputAsBase64ByNameRequest(ns1.getOutputAsBase64ByNameInput_Dec):
    if not hasattr(ns1.getOutputAsBase64ByNameInput_Dec(), "typecode"):
        typecode = ns1.getOutputAsBase64ByNameInput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.getOutputAsBase64ByNameInput_Dec.__init__(self, name=None, ns=None)


class getOutputAsBase64ByNameRequestWrapper(getOutputAsBase64ByNameRequest):
    """wrapper for document:literal message"""

    typecode = getOutputAsBase64ByNameRequest(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        getOutputAsBase64ByNameRequest.__init__(self, name=None, ns=None)


class getOutputAsBase64ByNameResponse(ns1.getOutputAsBase64ByNameOutput_Dec):
    if not hasattr(ns1.getOutputAsBase64ByNameOutput_Dec(), "typecode"):
        typecode = ns1.getOutputAsBase64ByNameOutput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.getOutputAsBase64ByNameOutput_Dec.__init__(self, name=None, ns=None)


class getOutputAsBase64ByNameResponseWrapper(getOutputAsBase64ByNameResponse):
    """wrapper for document:literal message"""

    typecode = getOutputAsBase64ByNameResponse(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        getOutputAsBase64ByNameResponse.__init__(self, name=None, ns=None)


class getOutputsRequest(ns1.getOutputsInput_Dec):
    if not hasattr(ns1.getOutputsInput_Dec(), "typecode"):
        typecode = ns1.getOutputsInput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.getOutputsInput_Dec.__init__(self, name=None, ns=None)


class getOutputsRequestWrapper(getOutputsRequest):
    """wrapper for document:literal message"""

    typecode = getOutputsRequest(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        getOutputsRequest.__init__(self, name=None, ns=None)


class getOutputsResponse(ns1.getOutputsOutput_Dec):
    if not hasattr(ns1.getOutputsOutput_Dec(), "typecode"):
        typecode = ns1.getOutputsOutput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.getOutputsOutput_Dec.__init__(self, name=None, ns=None)


class getOutputsResponseWrapper(getOutputsResponse):
    """wrapper for document:literal message"""

    typecode = getOutputsResponse(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        getOutputsResponse.__init__(self, name=None, ns=None)


class launchJobBlockingRequest(ns1.launchJobBlockingInput_Dec):
    if not hasattr(ns1.launchJobBlockingInput_Dec(), "typecode"):
        typecode = ns1.launchJobBlockingInput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.launchJobBlockingInput_Dec.__init__(self, name=None, ns=None)


class launchJobBlockingRequestWrapper(launchJobBlockingRequest):
    """wrapper for document:literal message"""

    typecode = launchJobBlockingRequest(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        launchJobBlockingRequest.__init__(self, name=None, ns=None)


class launchJobBlockingResponse(ns1.launchJobBlockingOutput_Dec):
    if not hasattr(ns1.launchJobBlockingOutput_Dec(), "typecode"):
        typecode = ns1.launchJobBlockingOutput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.launchJobBlockingOutput_Dec.__init__(self, name=None, ns=None)


class launchJobBlockingResponseWrapper(launchJobBlockingResponse):
    """wrapper for document:literal message"""

    typecode = launchJobBlockingResponse(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        launchJobBlockingResponse.__init__(self, name=None, ns=None)


class launchJobRequest(ns1.launchJobInput_Dec):
    if not hasattr(ns1.launchJobInput_Dec(), "typecode"):
        typecode = ns1.launchJobInput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.launchJobInput_Dec.__init__(self, name=None, ns=None)


class launchJobRequestWrapper(launchJobRequest):
    """wrapper for document:literal message"""

    typecode = launchJobRequest(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        launchJobRequest.__init__(self, name=None, ns=None)


class launchJobResponse(ns1.launchJobOutput_Dec):
    if not hasattr(ns1.launchJobOutput_Dec(), "typecode"):
        typecode = ns1.launchJobOutput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.launchJobOutput_Dec.__init__(self, name=None, ns=None)


class launchJobResponseWrapper(launchJobResponse):
    """wrapper for document:literal message"""

    typecode = launchJobResponse(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        launchJobResponse.__init__(self, name=None, ns=None)


class queryStatusRequest(ns1.queryStatusInput_Dec):
    if not hasattr(ns1.queryStatusInput_Dec(), "typecode"):
        typecode = ns1.queryStatusInput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.queryStatusInput_Dec.__init__(self, name=None, ns=None)


class queryStatusRequestWrapper(queryStatusRequest):
    """wrapper for document:literal message"""

    typecode = queryStatusRequest(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        queryStatusRequest.__init__(self, name=None, ns=None)


class queryStatusResponse(ns1.queryStatusOutput_Dec):
    if not hasattr(ns1.queryStatusOutput_Dec(), "typecode"):
        typecode = ns1.queryStatusOutput_Dec()

    def __init__(self, name=None, ns=None):
        ns1.queryStatusOutput_Dec.__init__(self, name=None, ns=None)


class queryStatusResponseWrapper(queryStatusResponse):
    """wrapper for document:literal message"""

    typecode = queryStatusResponse(name=None, ns=None).typecode

    def __init__(self, name=None, ns=None, **kw):
        queryStatusResponse.__init__(self, name=None, ns=None)



##################################################
# ./AppService_services_types.py
# generated by ZSI.wsdl2python
#
#
##################################################


import ZSI
from ZSI.TCcompound import Struct

##############################
# targetNamespace
#
# http://nbcr.sdsc.edu/opal/types
##############################


# imported as: ns1
class nbcr_sdsc_edu_opal_types:
    targetNamespace = "http://nbcr.sdsc.edu/opal/types"

    class InputFileType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "InputFileType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._name = None
            self._contents = None

            TClist = [
                ZSI.TC.String(pname="name", aname="_name"),
                ZSI.TC.Base64String(pname="contents", aname="_contents"),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                **kw,
            )

        def Get_name(self):
            return self._name

        def Set_name(self, _name):
            self._name = _name

        def Get_contents(self):
            return self._contents

        def Set_contents(self, _contents):
            self._contents = _contents

    class FaultType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "FaultType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._message = None

            TClist = [
                ZSI.TC.String(pname="message", aname="_message", optional=1),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                **kw,
            )

        def Get_message(self):
            return self._message

        def Set_message(self, _message):
            self._message = _message

    class FlagsType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "FlagsType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._id = None
            self._tag = None
            self._textDesc = None

            TClist = [
                ZSI.TC.String(pname="id", aname="_id"),
                ZSI.TC.String(pname="tag", aname="_tag"),
                ZSI.TC.String(pname="textDesc", aname="_textDesc", optional=1),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                **kw,
            )

        def Get_id(self):
            return self._id

        def Set_id(self, _id):
            self._id = _id

        def Get_tag(self):
            return self._tag

        def Set_tag(self, _tag):
            self._tag = _tag

        def Get_textDesc(self):
            return self._textDesc

        def Set_textDesc(self, _textDesc):
            self._textDesc = _textDesc

    class ImplicitParamsType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "ImplicitParamsType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._id = None
            self._name = None
            self._extension = None
            self._ioType = None
            self._required = None
            self._semanticType = None
            self._textDesc = None
            self._min = None
            self._max = None

            TClist = [
                ZSI.TC.String(pname="id", aname="_id"),
                ZSI.TC.String(pname="name", aname="_name", optional=1),
                ZSI.TC.String(pname="extension", aname="_extension", optional=1),
                ns1.IOType_Def(name="ioType", ns=ns),
                ZSI.TC.Boolean(pname="required", aname="_required", optional=1),
                ZSI.TC.String(pname="semanticType", aname="_semanticType", optional=1),
                ZSI.TC.String(pname="textDesc", aname="_textDesc", optional=1),
                ZSI.TCnumbers.Iint(pname="min", aname="_min", optional=1),
                ZSI.TCnumbers.Iint(pname="max", aname="_max", optional=1),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                **kw,
            )

        def Get_id(self):
            return self._id

        def Set_id(self, _id):
            self._id = _id

        def Get_name(self):
            return self._name

        def Set_name(self, _name):
            self._name = _name

        def Get_extension(self):
            return self._extension

        def Set_extension(self, _extension):
            self._extension = _extension

        def Get_ioType(self):
            return self._ioType

        def Set_ioType(self, _ioType):
            self._ioType = _ioType

        def Get_required(self):
            return self._required

        def Set_required(self, _required):
            self._required = _required

        def Get_semanticType(self):
            return self._semanticType

        def Set_semanticType(self, _semanticType):
            self._semanticType = _semanticType

        def Get_textDesc(self):
            return self._textDesc

        def Set_textDesc(self, _textDesc):
            self._textDesc = _textDesc

        def Get_min(self):
            return self._min

        def Set_min(self, _min):
            self._min = _min

        def Get_max(self):
            return self._max

        def Set_max(self, _max):
            self._max = _max

    class ParamsType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "ParamsType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._id = None
            self._tag = None
            self._paramType = None
            self._ioType = None
            self._required = None
            self._value = None
            self._semanticType = None
            self._textDesc = None

            TClist = [
                ZSI.TC.String(pname="id", aname="_id"),
                ZSI.TC.String(pname="tag", aname="_tag", optional=1),
                ns1.ParamType_Def(name="paramType", ns=ns),
                ns1.IOType_Def(name="ioType", ns=ns, optional=1),
                ZSI.TC.Boolean(pname="required", aname="_required", optional=1),
                ZSI.TC.String(pname="value", aname="_value", repeatable=1, optional=1),
                ZSI.TC.String(pname="semanticType", aname="_semanticType", optional=1),
                ZSI.TC.String(pname="textDesc", aname="_textDesc", optional=1),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                hasextras=1,
                **kw,
            )

        def Get_id(self):
            return self._id

        def Set_id(self, _id):
            self._id = _id

        def Get_tag(self):
            return self._tag

        def Set_tag(self, _tag):
            self._tag = _tag

        def Get_paramType(self):
            return self._paramType

        def Set_paramType(self, _paramType):
            self._paramType = _paramType

        def Get_ioType(self):
            return self._ioType

        def Set_ioType(self, _ioType):
            self._ioType = _ioType

        def Get_required(self):
            return self._required

        def Set_required(self, _required):
            self._required = _required

        def Get_value(self):
            return self._value

        def Set_value(self, _value):
            self._value = _value

        def Get_semanticType(self):
            return self._semanticType

        def Set_semanticType(self, _semanticType):
            self._semanticType = _semanticType

        def Get_textDesc(self):
            return self._textDesc

        def Set_textDesc(self, _textDesc):
            self._textDesc = _textDesc

    class OutputFileType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "OutputFileType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._name = None
            self._url = None

            TClist = [
                ZSI.TC.String(pname="name", aname="_name"),
                ZSI.TC.URI(pname="url", aname="_url"),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                **kw,
            )

        def Get_name(self):
            return self._name

        def Set_name(self, _name):
            self._name = _name

        def Get_url(self):
            return self._url

        def Set_url(self, _url):
            self._url = _url

    class AppMetadataInputType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "AppMetadataInputType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars

            TClist = []

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                **kw,
            )

    class StatusOutputType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "StatusOutputType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._code = None
            self._message = None
            self._baseURL = None

            TClist = [
                ZSI.TCnumbers.Iint(pname="code", aname="_code"),
                ZSI.TC.String(pname="message", aname="_message"),
                ZSI.TC.URI(pname="baseURL", aname="_baseURL"),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                **kw,
            )

        def Get_code(self):
            return self._code

        def Set_code(self, _code):
            self._code = _code

        def Get_message(self):
            return self._message

        def Set_message(self, _message):
            self._message = _message

        def Get_baseURL(self):
            return self._baseURL

        def Set_baseURL(self, _baseURL):
            self._baseURL = _baseURL

    class AppConfigInputType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "AppConfigInputType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars

            TClist = []

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                **kw,
            )

    class OutputsByNameInputType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "OutputsByNameInputType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._jobID = None
            self._fileName = None

            TClist = [
                ZSI.TC.String(pname="jobID", aname="_jobID"),
                ZSI.TC.String(pname="fileName", aname="_fileName"),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                **kw,
            )

        def Get_jobID(self):
            return self._jobID

        def Set_jobID(self, _jobID):
            self._jobID = _jobID

        def Get_fileName(self):
            return self._fileName

        def Set_fileName(self, _fileName):
            self._fileName = _fileName

    class getOutputAsBase64ByNameOutput_Dec(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/literals"
        literal = "getOutputAsBase64ByNameOutput"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema

            # internal vars
            self._item = None

            TClist = [
                ZSI.TCnumbers.Ibyte(
                    pname="item", aname="_item", repeatable=1, optional=1
                ),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                hasextras=1,
                **kw,
            )

        def Get_item(self):
            return self._item

        def Set_item(self, _item):
            self._item = _item

    class JobInputType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "JobInputType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._argList = None
            self._numProcs = None
            self._inputFile = None

            TClist = [
                ZSI.TC.String(pname="argList", aname="_argList", optional=1),
                ZSI.TCnumbers.Iint(pname="numProcs", aname="_numProcs", optional=1),
                ns1.InputFileType_Def(
                    name="inputFile", ns=ns, repeatable=1, optional=1
                ),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                hasextras=1,
                **kw,
            )

        def Get_argList(self):
            return self._argList

        def Set_argList(self, _argList):
            self._argList = _argList

        def Get_numProcs(self):
            return self._numProcs

        def Set_numProcs(self, _numProcs):
            self._numProcs = _numProcs

        def Get_inputFile(self):
            return self._inputFile

        def Set_inputFile(self, _inputFile):
            self._inputFile = _inputFile

    class opalFaultOutput_Dec(FaultType_Def):
        literal = "opalFaultOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema

            ns1.FaultType_Def.__init__(self, name=name, ns=ns, **kw)
            self.typecode = ns1.FaultType_Def(name=name, ns=ns, **kw)

    class FlagsArrayType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "FlagsArrayType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._flag = None

            TClist = [
                ns1.FlagsType_Def(name="flag", ns=ns, repeatable=1, optional=1),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                hasextras=1,
                **kw,
            )

        def Get_flag(self):
            return self._flag

        def Set_flag(self, _flag):
            self._flag = _flag

    class ImplicitParamsArrayType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "ImplicitParamsArrayType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._param = None

            TClist = [
                ns1.ImplicitParamsType_Def(
                    name="param", ns=ns, repeatable=1, optional=1
                ),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                hasextras=1,
                **kw,
            )

        def Get_param(self):
            return self._param

        def Set_param(self, _param):
            self._param = _param

    class ParamsArrayType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "ParamsArrayType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._separator = None
            self._param = None

            TClist = [
                ZSI.TC.String(pname="separator", aname="_separator", optional=1),
                ns1.ParamsType_Def(name="param", ns=ns, repeatable=1, optional=1),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                hasextras=1,
                **kw,
            )

        def Get_separator(self):
            return self._separator

        def Set_separator(self, _separator):
            self._separator = _separator

        def Get_param(self):
            return self._param

        def Set_param(self, _param):
            self._param = _param

    class JobOutputType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "JobOutputType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._stdOut = None
            self._stdErr = None
            self._outputFile = None

            TClist = [
                ZSI.TC.URI(pname="stdOut", aname="_stdOut", optional=1),
                ZSI.TC.URI(pname="stdErr", aname="_stdErr", optional=1),
                ns1.OutputFileType_Def(
                    name="outputFile", ns=ns, repeatable=1, optional=1
                ),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                hasextras=1,
                **kw,
            )

        def Get_stdOut(self):
            return self._stdOut

        def Set_stdOut(self, _stdOut):
            self._stdOut = _stdOut

        def Get_stdErr(self):
            return self._stdErr

        def Set_stdErr(self, _stdErr):
            self._stdErr = _stdErr

        def Get_outputFile(self):
            return self._outputFile

        def Set_outputFile(self, _outputFile):
            self._outputFile = _outputFile

    class getAppMetadataInput_Dec(AppMetadataInputType_Def):
        literal = "getAppMetadataInput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema

            ns1.AppMetadataInputType_Def.__init__(self, name=name, ns=ns, **kw)
            self.typecode = ns1.AppMetadataInputType_Def(name=name, ns=ns, **kw)

    class JobSubOutputType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "JobSubOutputType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._jobID = None
            self._status = None

            TClist = [
                ZSI.TC.String(pname="jobID", aname="_jobID"),
                ns1.StatusOutputType_Def(name="status", ns=ns),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                **kw,
            )

        def Get_jobID(self):
            return self._jobID

        def Set_jobID(self, _jobID):
            self._jobID = _jobID

        def Get_status(self):
            return self._status

        def Set_status(self, _status):
            self._status = _status

    class queryStatusOutput_Dec(StatusOutputType_Def):
        literal = "queryStatusOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema

            ns1.StatusOutputType_Def.__init__(self, name=name, ns=ns, **kw)
            self.typecode = ns1.StatusOutputType_Def(name=name, ns=ns, **kw)

    class destroyOutput_Dec(StatusOutputType_Def):
        literal = "destroyOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema

            ns1.StatusOutputType_Def.__init__(self, name=name, ns=ns, **kw)
            self.typecode = ns1.StatusOutputType_Def(name=name, ns=ns, **kw)

    class getAppConfigInput_Dec(AppConfigInputType_Def):
        literal = "getAppConfigInput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema

            ns1.AppConfigInputType_Def.__init__(self, name=name, ns=ns, **kw)
            self.typecode = ns1.AppConfigInputType_Def(name=name, ns=ns, **kw)

    class getOutputAsBase64ByNameInput_Dec(OutputsByNameInputType_Def):
        literal = "getOutputAsBase64ByNameInput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema

            ns1.OutputsByNameInputType_Def.__init__(self, name=name, ns=ns, **kw)
            self.typecode = ns1.OutputsByNameInputType_Def(name=name, ns=ns, **kw)

    class launchJobInput_Dec(JobInputType_Def):
        literal = "launchJobInput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema

            ns1.JobInputType_Def.__init__(self, name=name, ns=ns, **kw)
            self.typecode = ns1.JobInputType_Def(name=name, ns=ns, **kw)

    class launchJobBlockingInput_Dec(JobInputType_Def):
        literal = "launchJobBlockingInput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema

            ns1.JobInputType_Def.__init__(self, name=name, ns=ns, **kw)
            self.typecode = ns1.JobInputType_Def(name=name, ns=ns, **kw)

    class ArgumentsType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "ArgumentsType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._flags = None
            self._taggedParams = None
            self._untaggedParams = None
            self._implicitParams = None

            TClist = [
                ns1.FlagsArrayType_Def(name="flags", ns=ns, optional=1),
                ns1.ParamsArrayType_Def(name="taggedParams", ns=ns, optional=1),
                ns1.ParamsArrayType_Def(name="untaggedParams", ns=ns, optional=1),
                ns1.ImplicitParamsArrayType_Def(
                    name="implicitParams", ns=ns, optional=1
                ),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                **kw,
            )

        def Get_flags(self):
            return self._flags

        def Set_flags(self, _flags):
            self._flags = _flags

        def Get_taggedParams(self):
            return self._taggedParams

        def Set_taggedParams(self, _taggedParams):
            self._taggedParams = _taggedParams

        def Get_untaggedParams(self):
            return self._untaggedParams

        def Set_untaggedParams(self, _untaggedParams):
            self._untaggedParams = _untaggedParams

        def Get_implicitParams(self):
            return self._implicitParams

        def Set_implicitParams(self, _implicitParams):
            self._implicitParams = _implicitParams

    class BlockingOutputType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "BlockingOutputType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._status = None
            self._jobOut = None

            TClist = [
                ns1.StatusOutputType_Def(name="status", ns=ns),
                ns1.JobOutputType_Def(name="jobOut", ns=ns),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                **kw,
            )

        def Get_status(self):
            return self._status

        def Set_status(self, _status):
            self._status = _status

        def Get_jobOut(self):
            return self._jobOut

        def Set_jobOut(self, _jobOut):
            self._jobOut = _jobOut

    class getOutputsOutput_Dec(JobOutputType_Def):
        literal = "getOutputsOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema

            ns1.JobOutputType_Def.__init__(self, name=name, ns=ns, **kw)
            self.typecode = ns1.JobOutputType_Def(name=name, ns=ns, **kw)

    class launchJobOutput_Dec(JobSubOutputType_Def):
        literal = "launchJobOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema

            ns1.JobSubOutputType_Def.__init__(self, name=name, ns=ns, **kw)
            self.typecode = ns1.JobSubOutputType_Def(name=name, ns=ns, **kw)

    class AppMetadataType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "AppMetadataType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._usage = None
            self._info = None
            self._types = None

            TClist = [
                ZSI.TC.String(pname="usage", aname="_usage"),
                ZSI.TC.String(pname="info", aname="_info", repeatable=1, optional=1),
                ns1.ArgumentsType_Def(name="types", ns=ns, optional=1),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                hasextras=1,
                **kw,
            )

        def Get_usage(self):
            return self._usage

        def Set_usage(self, _usage):
            self._usage = _usage

        def Get_info(self):
            return self._info

        def Set_info(self, _info):
            self._info = _info

        def Get_types(self):
            return self._types

        def Set_types(self, _types):
            self._types = _types

    class launchJobBlockingOutput_Dec(BlockingOutputType_Def):
        literal = "launchJobBlockingOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema

            ns1.BlockingOutputType_Def.__init__(self, name=name, ns=ns, **kw)
            self.typecode = ns1.BlockingOutputType_Def(name=name, ns=ns, **kw)

    class AppConfigType_Def(ZSI.TCcompound.Struct):
        schema = "http://nbcr.sdsc.edu/opal/types"
        type = "AppConfigType"

        def __init__(self, name=None, ns=None, **kw):
            # internal vars
            self._metadata = None
            self._binaryLocation = None
            self._defaultArgs = None
            self._parallel = None

            TClist = [
                ns1.AppMetadataType_Def(name="metadata", ns=ns),
                ZSI.TC.String(pname="binaryLocation", aname="_binaryLocation"),
                ZSI.TC.String(pname="defaultArgs", aname="_defaultArgs", optional=1),
                ZSI.TC.Boolean(pname="parallel", aname="_parallel"),
            ]

            oname = name

            if name:
                aname = "_%s" % name
                if ns:
                    oname += ' xmlns="%s"' % ns
                else:
                    oname += ' xmlns="%s"' % self.__class__.schema
            else:
                aname = None

            ZSI.TCcompound.Struct.__init__(
                self,
                self.__class__,
                TClist,
                pname=name,
                inorder=0,
                aname=aname,
                oname=oname,
                **kw,
            )

        def Get_metadata(self):
            return self._metadata

        def Set_metadata(self, _metadata):
            self._metadata = _metadata

        def Get_binaryLocation(self):
            return self._binaryLocation

        def Set_binaryLocation(self, _binaryLocation):
            self._binaryLocation = _binaryLocation

        def Get_defaultArgs(self):
            return self._defaultArgs

        def Set_defaultArgs(self, _defaultArgs):
            self._defaultArgs = _defaultArgs

        def Get_parallel(self):
            return self._parallel

        def Set_parallel(self, _parallel):
            self._parallel = _parallel

    class getAppMetadataOutput_Dec(AppMetadataType_Def):
        literal = "getAppMetadataOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema

            ns1.AppMetadataType_Def.__init__(self, name=name, ns=ns, **kw)
            self.typecode = ns1.AppMetadataType_Def(name=name, ns=ns, **kw)

    class getAppConfigOutput_Dec(AppConfigType_Def):
        literal = "getAppConfigOutput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema

            ns1.AppConfigType_Def.__init__(self, name=name, ns=ns, **kw)
            self.typecode = ns1.AppConfigType_Def(name=name, ns=ns, **kw)

    class IOType_Def(ZSI.TC.String):
        tag = "tns:IOType"

        def __init__(self, name=None, ns=None, **kw):
            aname = None
            if name:
                kw["pname"] = name
                kw["aname"] = "_%s" % name
                kw["oname"] = '%s xmlns:tns="%s"' % (name, ns1.targetNamespace)
            ZSI.TC.String.__init__(self, **kw)

    class ParamType_Def(ZSI.TC.String):
        tag = "tns:ParamType"

        def __init__(self, name=None, ns=None, **kw):
            aname = None
            if name:
                kw["pname"] = name
                kw["aname"] = "_%s" % name
                kw["oname"] = '%s xmlns:tns="%s"' % (name, ns1.targetNamespace)
            ZSI.TC.String.__init__(self, **kw)

    class destroyInput_Dec(ZSI.TC.String):
        literal = "destroyInput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema
            kw["oname"] = '%s xmlns="%s"' % (name, ns)

            ZSI.TC.String.__init__(self, pname=name, aname="%s" % name, **kw)

    class queryStatusInput_Dec(ZSI.TC.String):
        literal = "queryStatusInput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema
            kw["oname"] = '%s xmlns="%s"' % (name, ns)

            ZSI.TC.String.__init__(self, pname=name, aname="%s" % name, **kw)

    class getOutputsInput_Dec(ZSI.TC.String):
        literal = "getOutputsInput"
        schema = "http://nbcr.sdsc.edu/opal/types"

        def __init__(self, name=None, ns=None, **kw):
            name = name or self.__class__.literal
            ns = ns or self.__class__.schema
            kw["oname"] = '%s xmlns="%s"' % (name, ns)

            ZSI.TC.String.__init__(self, pname=name, aname="%s" % name, **kw)


# define class alias for subsequent ns classes
ns1 = nbcr_sdsc_edu_opal_types



"""
    Structures for PDB2PQR

    This module contains the structure objects used in PDB2PQR and their
    associated methods.

    ----------------------------
   
    PDB2PQR -- An automated pipeline for the setup, execution, and analysis of
    Poisson-Boltzmann electrostatics calculations

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin; 
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific 
    Northwest National Laboratory, operated by Battelle Memorial Institute, 
    Pacific Northwest Division for the U.S. Department Energy.; 
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, 
	are permitted provided that the following conditions are met:

		* Redistributions of source code must retain the above copyright notice, 
		  this list of conditions and the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, 
		  this list of conditions and the following disclaimer in the documentation 
		  and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
	OF THE POSSIBILITY OF SUCH DAMAGE.

    ----------------------------

"""

__date__ = "28 February 2006"
__author__ = "Todd Dolinsky"

BACKBONE = ["N", "CA", "C", "O", "O2", "HA", "HN", "H", "tN"]

import string
from pdb import *
from utilities import *
from quatfit import *
from errors import PDBInternalError


class Chain:
    """
    Chain class

    The chain class contains information about each chain within a given
    Protein object.
    """

    def __init__(self, chainID):
        """
        Initialize the class

        Parameters
            chainID: The chainID for this chain as denoted in
                     the PDB file (string)
        """

        self.chainID = chainID
        self.residues = []

    def get(self, name):
        """
        Get a member of the Chain class

        Parameters
            name:     The name of the member
        Possible Values
            ID:       The ID of the chain
            Residues: The list of residues within the Chain
        Returns
            item:     The value of the member
        """
        if name == "atoms":
            self.getAtoms()
        else:
            try:
                item = getattr(self, name)
                return item
            except AttributeError:
                message = 'Unable to get object "%s" in class Chain' % name
                raise PDBInternalError(message)

    def addResidue(self, residue):
        """
        Add a residue to the chain

        Parameters
            residue: The residue to be added (Residue)
        """
        self.residues.append(residue)

    def numResidues(self):
        """
        Get the number of residues for the chain

        Returns
            count:  Number of residues in the chain (int)
        """
        count = 0
        for residue in self.residues:
            count += 1
        return count

    def renumberResidues(self):
        """
        Renumber Atoms based on actual Residue number and not PDB resSeq
        """
        count = 1
        for residue in self.residues:
            residue.setResSeq(count)
            count += 1

    def numAtoms(self):
        """
        Get the number of atoms for the chain

        Returns
            count:  Number of atoms in the chain (int)
        """
        count = len(self.getAtoms())
        return count

    def getResidues(self):
        """
        Return a list of Residue objects in this chain
        """
        return self.residues

    def getAtoms(self):
        """
        Return a list of Atom objects contained in this chain

        Returns
            atomlist: List of Atom objects (list)
        """
        atomlist = []
        for residue in self.residues:
            myList = residue.get("atoms")
            for atom in myList:
                atomlist.append(atom)
        return atomlist

    def getSummary(self):
        output = []
        for residue in self.residues:
            output.append(residue.letterCode())

        return "".join(output)


class Residue:
    """
    Residue class

    The residue class contains a list of Atom objects associated with that
    residue and other helper functions.
    """

    def __init__(self, atoms):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """

        sampleAtom = atoms[-1]

        self.atoms = []
        self.name = sampleAtom.resName
        self.chainID = sampleAtom.chainID
        self.resSeq = sampleAtom.resSeq
        self.iCode = sampleAtom.iCode
        #
        #
        self.map = {}

        self.naname = None

        atomclass = ""
        for a in atoms:
            if isinstance(a, ATOM):
                atomclass = "ATOM"
            elif isinstance(a, HETATM):
                atomclass = "HETATM"
            atom = Atom(a, atomclass, self)
            atomname = atom.get("name")
            if atomname not in self.map:
                self.addAtom(atom)
            else:  # Don't add duplicate atom
                oldatom = self.getAtom(atomname)
                oldatom.set("altLoc", "")

        if self.name == "HOH":
            self.name = "WAT"
            for atom in self.atoms:
                atom.set("resName", "WAT")

    def __str__(self):
        """
        Basic string representation
        """
        text = "%s %s %i%s" % (self.name, self.chainID, self.resSeq, self.iCode)
        return text

    # TODO: Kill this in a fire.
    def get(self, name):
        """
        Get a member of the Residue class

        Parameters
            name:          The name of the member (string)
        Possible Values
            atoms:         The atoms in the residue
            name:          The name of the residue
            chainID:       The chainID associated with the residue
            resSeq:        The sequence number of the residue
            icode:         The iCode of the residue
            SSbonded:      1 if the residue has a SS bond, 0 otherwise
            SSbondpartner: The residue of the bond partner
            type:          The type associated with this residue
            isNterm:       # of hydrogens if the residue is the N-Terminus, 0 otherwise
            isCterm:       1 if the residue is the C-Terminus, 0 otherwise
            missing:     List of missing atoms of the residue
        Returns
            item:          The value of the member
        """
        try:
            item = getattr(self, name)
            return item
        except AttributeError:
            message = 'Unable to access object "%s" in class Residue' % name
            raise PDBInternalError(message)

    def set(self, name, value):
        """
        Set a member of the Residue class to a specific value

        Parameters
            name:          The name of the object to set (string)
            value:         The object to append
        Possible Values
            atoms:         The atoms in the residue
            name:          The name of the residue
            chain:         The chainID associated with the residue
            resSeq:        The sequence number of the residue
            icode:         The iCode of the residue
            SSbonded:      1 if the residue has a SS bond, 0 otherwise
            SSbondpartner: The residue of the bond partner
            type:          The type associated with this residue
            isNterm:       # of hydrogens if the residue is the N-Terminus, 0 otherwise
            isCterm:       1 if the residue is the C-Terminus, 0 otherwise
            isDirty:       1 if the residue is not missing atoms,
                           0 otherwise
        Notes
            resSeq points to the residue.setResSeq function
        Returns
            item:          The value of the member
        """
        if name == "resSeq":
            self.setResSeq(value)
        else:
            try:
                setattr(self, name, value)
            except AttributeError:
                message = 'Unable to set object "%s" in class Residue' % name
                raise PDBInternalError(message)

    def update_terminus_status(self):
        """Update the isNterms and isCterm flags"""
        #
        # If Nterm then update counter of hydrogens
        #
        if self.isNterm:
            count = 0
            atoms = ["H", "H2", "H3"]
            for atom in atoms:
                for atom2 in self.atoms:
                    atomname = atom2.get("name")
                    if atom == atomname:
                        count = count + 1
            self.isNterm = count
        #
        # If Cterm then update counter
        #
        if self.isCterm:
            self.isCterm = None
            for atom in self.atoms:
                atomname = atom.get("name")
                if atomname == "HO":
                    self.isCterm = 2
                    break
            if not self.isCterm:
                self.isCterm = 1
        return

    def numAtoms(self):
        """
        Get the number of atoms for the residue

        Returns
            Number of atoms in the residue (int)
        """
        return len(self.atoms)

    def setResSeq(self, value):
        """
        Set the atom field resSeq to a certain value and
        change the residue's information.  The icode field is no longer
        useful.

        Parameters
            value:  The new value of resSeq (int)
        """
        self.iCode = ""
        self.resSeq = value
        for atom in self.atoms:
            atom.set("resSeq", value)
            # atom.set("iCode","")

    def setChainID(self, value):
        """
        Set the chainID field to a certain value
        """
        self.chainID = value
        for atom in self.atoms:
            atom.set("chainID", value)

    def addAtom(self, atom):
        """
        Add the atom object to the residue.

        Parameters
            atom: The object to be added (ATOM)
        """
        self.atoms.append(atom)
        self.map[atom.get("name")] = atom

    def removeAtom(self, atomname):
        """
        Remove an atom from the residue object.

        Parameters
            atomname: The name of the atom to be removed (string)
        """

        # Delete the atom from the map

        atom = self.map[atomname]
        bonds = atom.bonds
        del self.map[atomname]

        # Delete the atom from the list

        self.atoms.remove(atom)

        # Delete all instances of the atom as a bond

        for bondatom in bonds:
            if atom in bondatom.bonds:
                bondatom.bonds.remove(atom)

        del atom

    def renameAtom(self, oldname, newname):
        """
        Rename an atom to a new name

        Parameters
            oldname: The old atom name (string)
            newname: The new atom name (string)
        """
        atom = self.map[oldname]
        atom.set("name", newname)
        self.map[newname] = atom
        del self.map[oldname]

    def createAtom(self, name, newcoords, type):
        """
        Add a new atom object to the residue. Uses an atom
        currently in the residue to seed the new atom
        object, then replaces the coordinates and name accordingly.

        Parameters
            name:      The name of the new atom (string)
            newcoords: The x,y,z coordinates of the new atom (list)
            type:      The type of atom, ATOM or HETATM
        """
        oldatom = self.atoms[0]
        newatom = Atom(oldatom, type, self)
        newatom.set("x", newcoords[0])
        newatom.set("y", newcoords[1])
        newatom.set("z", newcoords[2])
        newatom.set("name", name)
        newatom.set("occupancy", 1.00)
        newatom.set("tempFactor", 0.00)
        self.addAtom(newatom)

    def addMissing(self, value):
        """
        Add the value to the list of missing atoms

        Parameters
            value: The name of the missing atom (string)
        """
        self.missing.append(value)

    def getAtom(self, name):
        """
        Retrieve an atom from the mapping

        Parameters
            resname: The name of the residue to retrieve (string)
        """
        return self.map.get(name)

    def getAtoms(self):
        return self.atoms

    def hasAtom(self, name):
        return name in self.map

    def getCharge(self):
        """
        Get the total charge of the residue.  In order to get rid
        of floating point rounding error, do the string
        transformation.

        Returns:
            charge: The charge of the residue (float)
        """
        # ffcharge can be None.
        charge = (atom.ffcharge for atom in self.atoms if atom.ffcharge)
        charge = sum(charge)
        #        charge = 0.0
        #        for atom in self.atoms:
        #            atomcharge = atom.get("ffcharge")
        #            if atomcharge != None:
        #                charge += atomcharge

        charge = float("%.4f" % charge)
        return charge

    def renameResidue(self, name):
        """
        Rename a given residue

        Parameters
            name:       The new name of the residue
        """
        self.name = name
        for atom in self.atoms:
            atom.resName = name

    def rotateTetrahedral(self, atom1, atom2, angle):
        """
        Rotate about the atom1-atom2 bond by a given angle
        All atoms connected to atom2 will rotate.

        Parameters:
            atom1:  The first atom of the bond to rotate about (atom)
            atom2:  The second atom of the bond to rotate about (atom)
            angle:  The number of degrees to rotate (float)
        """
        moveatoms = []
        movecoords = []

        initcoords = subtract(atom2.getCoords(), atom1.getCoords())

        # Determine which atoms to rotate

        for atom in atom2.bonds:
            if atom == atom1:
                continue
            moveatoms.append(atom)
            movecoords.append(subtract(atom.getCoords(), atom1.getCoords()))

        newcoords = qchichange(initcoords, movecoords, angle)
        for i in range(len(moveatoms)):
            atom = moveatoms[i]
            x = newcoords[i][0] + atom1.get("x")
            y = newcoords[i][1] + atom1.get("y")
            z = newcoords[i][2] + atom1.get("z")
            atom.set("x", x)
            atom.set("y", y)
            atom.set("z", z)

    def setDonorsAndAcceptors(self):
        """
        Set the donors and acceptors within the residue
        """
        if not hasattr(self, "reference"):
            return
        for atom in self.getAtoms():
            atomname = atom.get("name")
            resname = self.name

            atom.set("hdonor", 0)
            atom.set("hacceptor", 0)

            if atomname.startswith("N"):
                bonded = 0
                for bondedatom in atom.bonds:
                    if bondedatom.isHydrogen():
                        atom.set("hdonor", 1)
                        bonded = 1
                        break
                if not bonded and self.reference.name == "HIS":
                    atom.set("hacceptor", 1)

            elif atomname.startswith("O") or (
                atomname.startswith("S") and self.reference.name == "CYS"
            ):
                atom.set("hacceptor", 1)
                for bondedatom in atom.bonds:
                    if bondedatom.isHydrogen():
                        atom.set("hdonor", 1)
                        break

    def reorder(self):
        """
        Reorder the atoms to start with N, CA, C, O if they exist
        """
        templist = []
        if self.hasAtom("N"):
            templist.append(self.getAtom("N"))
        if self.hasAtom("CA"):
            templist.append(self.getAtom("CA"))
        if self.hasAtom("C"):
            templist.append(self.getAtom("C"))
        if self.hasAtom("O"):
            templist.append(self.getAtom("O"))

        # Add remaining atoms
        for atom in self.atoms:
            if atom.name not in ["N", "CA", "C", "O"]:
                templist.append(atom)

        # Change the list pointer

        self.atoms = templist[:]

    def letterCode(self):
        return "X"


class Atom(ATOM):
    """
    Class Atom

    The Atom class inherits off the ATOM object in pdb.py.  It is used
    for adding fields not found in the pdb that may be useful for analysis.
    Also simplifies code by combining ATOM and HETATM objects into a
    single class.
    """

    def __init__(self, atom, type, residue):
        """
        Initialize the new Atom object by using the old object.

        Parameters
            atom:    The original ATOM object (ATOM)
            type:    Either ATOM or HETATM (string)
            residue: A pointer back to the parent residue object (Residue)
        """
        if type == "ATOM" or type == "HETATM":
            self.type = type
        else:
            raise PDBInternalError(
                "Invalid atom type %s (Atom Class IN structures.py)!" % type
            )
        self.serial = atom.serial
        self.name = atom.name
        self.altLoc = atom.altLoc
        self.resName = atom.resName
        self.chainID = atom.chainID
        self.resSeq = atom.resSeq
        self.iCode = atom.iCode
        self.x = atom.x
        self.y = atom.y
        self.z = atom.z
        self.occupancy = atom.occupancy
        self.tempFactor = atom.tempFactor
        self.segID = atom.segID
        self.element = atom.element
        self.charge = atom.charge

        self.bonds = []
        self.reference = None
        self.residue = residue
        self.radius = None
        self.ffcharge = None
        self.hdonor = 0
        self.hacceptor = 0
        self.cell = None
        self.added = 0
        self.optimizeable = 0
        self.refdistance = 0
        self.id = None

        self.mol2charge = None
        if hasattr(atom, "mol2charge"):
            self.mol2charge = atom.mol2charge

    def getCommonStringRep(self, chainflag=False):
        """
        Returns a string of the common column of the new atom type.
        Uses the ATOM string output but changes the first field
        to either by ATOM or HETATM as necessary.

        This is used to create the output for pqr and pdb files! No touchy!

        Returns
            outstr: String with ATOM/HETATM field set appropriately
        """
        outstr = ""
        tstr = self.type
        outstr += string.ljust(tstr, 6)[:6]
        tstr = "%d" % self.serial
        outstr += string.rjust(tstr, 5)[:5]
        outstr += " "
        tstr = self.name
        if len(tstr) == 4 or len(tstr.strip("FLIP")) == 4:
            outstr += string.ljust(tstr, 4)[:4]
        else:
            outstr += " " + string.ljust(tstr, 3)[:3]

        tstr = self.resName
        if len(tstr) == 4:
            outstr += string.ljust(tstr, 4)[:4]
        else:
            outstr += " " + string.ljust(tstr, 3)[:3]

        outstr += " "
        if chainflag:
            tstr = self.chainID
        else:
            tstr = ""
        outstr += string.ljust(tstr, 1)[:1]
        tstr = "%d" % self.resSeq
        outstr += string.rjust(tstr, 4)[:4]
        if self.iCode != "":
            outstr += "%s   " % self.iCode
        else:
            outstr += "    "
        tstr = "%8.3f" % self.x
        outstr += string.ljust(tstr, 8)[:8]
        tstr = "%8.3f" % self.y
        outstr += string.ljust(tstr, 8)[:8]
        tstr = "%8.3f" % self.z
        outstr += string.ljust(tstr, 8)[:8]
        return outstr

    def __str__(self):
        """
        Returns a string of the new atom type.  Uses the ATOM string
        output but changes the first field to either by ATOM or
        HETATM as necessary.

        This is used to create the output for pqr files! No touchy!

        Returns
            str: String with ATOM/HETATM field set appropriately
        """
        return self.getPQRString()

    def getPQRString(self, chainflag=False):
        """
        Returns a string of the new atom type.  Uses the ATOM string
        output but changes the first field to either by ATOM or
        HETATM as necessary.

        This is used to create the output for pqr files! No touchy!

        Returns
            str: String with ATOM/HETATM field set appropriately
        """
        outstr = self.getCommonStringRep(chainflag=chainflag)

        if self.ffcharge != None:
            ffcharge = "%.4f" % self.ffcharge
        else:
            ffcharge = "0.0000"
        outstr += string.rjust(ffcharge, 8)[:8]
        if self.radius != None:
            ffradius = "%.4f" % self.radius
        else:
            ffradius = "0.0000"
        outstr += string.rjust(ffradius, 7)[:7]

        return outstr

    def getPDBString(self):
        """
        Returns a string of the new atom type.  Uses the ATOM string
        output but changes the first field to either by ATOM or
        HETATM as necessary.

        This is for the pdb representation of the atom. The propka30 module
        depends on this being correct. No touchy!

        Returns
            str: String with ATOM/HETATM field set appropriately
        """
        outstr = self.getCommonStringRep(chainflag=True)

        tstr = "%6.2f" % self.occupancy
        outstr += string.ljust(tstr, 6)[:6]
        tstr = "%6.2f" % self.tempFactor
        outstr += string.rjust(tstr, 6)[:6]
        # padding between temp factor and segID
        outstr += " " * 7
        tstr = self.segID
        outstr += string.ljust(tstr, 4)[:4]
        tstr = self.element
        outstr += string.ljust(tstr, 2)[:2]
        tstr = str(self.charge)
        outstr += string.ljust(tstr, 2)[:2]

        return outstr

    # TODO: What? Why? Isn't this Python?
    # Are we really doing attribute access based
    # on dynamic names? A search of the code says no.
    def get(self, name):
        """
        Get a member of the Atom class

        Parameters
            name:       The name of the member (string)
        Possible Values
            type:       The type of Atom (either ATOM or HETATM)
            serial:     Atom serial number
            name:       Atom name
            altLoc:     Alternate location
            resName:    Residue name
            chainID:    Chain identifier
            resSeq:     Residue sequence number
            iCode:      Code for insertion of residues
            x:          Orthogonal coordinates for X in Angstroms.
            y:          Orthogonal coordinates for Y in Angstroms.
            z:          Orthogonal coordinates for Z in Angstroms.
            occupancy:  Occupancy
            tempFactor: Temperature Factor
            segID:      Segment identifier
            element:    Element symbol
            charge:     Charge on the atom
            bonds:      The bonds associated with the atom
            interbonds: The intrabonds associated with the atom
            extrabonds: The extrabonds assocaited with the atom
            residue:    The parent residue of the atom
            radius:     The radius of the atom
            ffcharge:   The forcefield charge on the atom
            hdonor:     Whether the atom is a hydrogen donor
            hacceptor:  Whether the atom is a hydrogen acceptor
        Returns
            item:       The value of the member
        """
        try:
            item = getattr(self, name)
            return item
        except AttributeError:
            message = 'Unable to access object "%s" in class Atom' % name
            raise PDBInternalError(message)

    def set(self, name, value):
        """
        Set a member of the Atom class

        Parameters
            name:       The name of the member (string)
            value:      The value to set the member to
        Possible Values
            type:       The type of Atom (either ATOM or HETATM)
            serial:     Atom serial number
            name:       Atom name
            altLoc:     Alternate location
            resName:    Residue name
            chainID:    Chain identifier
            resSeq:     Residue sequence number
            iCode:      Code for insertion of residues
            x:          Orthogonal coordinates for X in Angstroms.
            y:          Orthogonal coordinates for Y in Angstroms.
            z:          Orthogonal coordinates for Z in Angstroms.
            occupancy:  Occupancy
            tempFactor: Temperature Factor
            segID:      Segment identifier
            element:    Element symbol
            charge:     Charge on the atom
            residue:    The parent residue of the atom
            radius:     The radius of the atom
            ffcharge:   The forcefield charge on the atom
            hdonor:     Whether the atom is a hydrogen donor
            hacceptor:  Whether the atom is a hydrogen acceptor
        """
        try:
            setattr(self, name, value)
        except AttributeError:
            message = 'Unable to set object "%s" in class Atom' % name
            raise PDBInternalError(message)

    def getCoords(self):
        """
        Return the x,y,z coordinates of the atom in list form

        Returns
            List of the coordinates (list)
        """
        return [self.x, self.y, self.z]

    def addBond(self, bondedatom):
        """
        Add a bond to the list of bonds

        Parameters:
            bondedatom: The atom to bond to (Atom)
        """
        self.bonds.append(bondedatom)

    def isHydrogen(self):
        """
        Is this atom a Hydrogen atom?

        Returns
            value: 1 if Atom is a Hydrogen, 0 otherwise
        """
        return self.name[0] == "H"

    def isBackbone(self):
        """
        Return true if atom name is in backbone, otherwise false

        Returns
            state: 1 if true, 0 if false
        """
        return self.name in BACKBONE

    def hasReference(self):
        """
        Determine if the atom object has a reference object or not.
        All known atoms should have reference objects.

        Returns
            True if atom has a reference object, False otherwise.
        """

        return self.reference != None



"""
    Amino Acid Structures for PDB2PQR

    This module contains the base amino acid structures for
    pdb2pqr.

    ----------------------------

    PDB2PQR -- An automated pipeline for the setup, execution, and analysis of
    Poisson-Boltzmann electrostatics calculations

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin;
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific
    Northwest National Laboratory, operated by Battelle Memorial Institute,
    Pacific Northwest Division for the U.S. Department Energy.;
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

		* Redistributions of source code must retain the above copyright notice,
		  this list of conditions and the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice,
		  this list of conditions and the following disclaimer in the documentation
		  and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
	OF THE POSSIBILITY OF SUCH DAMAGE.

    ----------------------------

"""

__date__ = "28 December 2006"
__author__ = "Todd Dolinsky"

import string
from structures import *
from src.errors import PDBInputError


class Amino(Residue):
    """
    Amino class

    This class provides standard features of the amino acids listed
    below

    Parameters
        atoms:  A list of Atom objects to be stored in this class
                 (list)
        ref:    The reference object for the amino acid.  Used to
                convert from the alternate naming scheme to the
                main naming scheme.
    """

    def __init__(self, atoms, ref):
        sampleAtom = atoms[-1]

        self.atoms = []
        self.name = sampleAtom.resName
        self.chainID = sampleAtom.chainID
        self.resSeq = sampleAtom.resSeq
        self.iCode = sampleAtom.iCode

        self.ffname = self.name
        self.map = {}
        self.dihedrals = []
        self.patches = []
        self.peptideC = None
        self.peptideN = None
        self.isNterm = 0
        self.isCterm = 0
        self.is5term = 0
        self.is3term = 0
        self.missing = []
        self.reference = ref
        self.fixed = 0
        self.stateboolean = {}

        # Create each atom

        for a in atoms:
            if a.name in ref.altnames:  # Rename atoms
                a.name = ref.altnames[a.name]

            if a.name not in self.map:
                atom = Atom(a, "ATOM", self)
                self.addAtom(atom)

    def createAtom(self, atomname, newcoords):
        """
        Create an atom.  Override the generic residue's version of
        createAtom().

        Parameters
            atomname:  The name of the atom (string)
            newcoords: The coordinates of the atom (list).
        """
        oldatom = self.atoms[0]
        newatom = Atom(oldatom, "ATOM", self)
        newatom.set("x", newcoords[0])
        newatom.set("y", newcoords[1])
        newatom.set("z", newcoords[2])
        newatom.set("name", atomname)
        newatom.set("occupancy", 1.00)
        newatom.set("tempFactor", 0.00)
        newatom.added = 1
        self.addAtom(newatom)

    def addAtom(self, atom):
        """
        Override the existing addAtom - include the link to the
        reference object
        """
        self.atoms.append(atom)
        atomname = atom.get("name")
        self.map[atomname] = atom
        try:
            atom.reference = self.reference.map[atomname]
            for bond in atom.reference.bonds:
                if self.hasAtom(bond):
                    bondatom = self.map[bond]
                    if bondatom not in atom.bonds:
                        atom.bonds.append(bondatom)
                    if atom not in bondatom.bonds:
                        bondatom.bonds.append(atom)
        except KeyError:
            atom.reference = None

    def addDihedralAngle(self, value):
        """
        Add the value to the list of chiangles

        Parameters
            value: The value to be added (float)
        """
        self.dihedrals.append(value)

    def setState(self):
        """
        Set the name to use for the forcefield based on the current
        state.  Uses N* and C* for termini.
        """
        if self.isNterm:
            if "NEUTRAL-NTERM" in self.patches:
                self.ffname = "NEUTRAL-N%s" % self.ffname
            else:
                self.ffname = "N%s" % self.ffname
        elif self.isCterm:
            if "NEUTRAL-CTERM" in self.patches:
                self.ffname = "NEUTRAL-C%s" % self.ffname
            else:
                self.ffname = "C%s" % self.ffname
        return


class ALA(Amino):
    """
    Alanine class

    This class gives data about the Alanine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "A"


class ARG(Amino):
    """
    Arginine class

    This class gives data about the Arginine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "R"

    def setState(self):
        """
        Set the name to use for the forcefield based on the current
        state.
        """
        if "AR0" in self.patches or self.name == "AR0":
            self.ffname = "AR0"
        Amino.setState(self)


class ASN(Amino):
    """
    Asparagine class

    This class gives data about the Asparagine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "N"


class ASP(Amino):
    """
    Aspartic Acid class

    This class gives data about the Aspartic Acid object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "D"

    def setState(self):
        """
        Set the name to use for the forcefield based on the current
        state.
        """
        if "ASH" in self.patches or self.name == "ASH":
            self.ffname = "ASH"
        Amino.setState(self)


class CYS(Amino):
    """
    Cysteine class

    This class gives data about the Cysteine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref
        self.SSbonded = 0
        self.SSbondedpartner = None

    def letterCode(self):
        return "C"

    def setState(self):
        """
        Set the state of the CYS object.  If SS-bonded, use CYX.  If
        negatively charged, use CYM.  If HG is not present, use CYX.
        """
        if "CYX" in self.patches or self.name == "CYX":
            self.ffname = "CYX"
        elif self.SSbonded:
            self.ffname = "CYX"
        elif "CYM" in self.patches or self.name == "CYM":
            self.ffname = "CYM"
        elif not self.hasAtom("HG"):
            self.ffname = "CYX"
        Amino.setState(self)


class GLN(Amino):
    """
    Glutamine class

    This class gives data about the Glutamine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "Q"


class GLU(Amino):
    """
    Glutamic Acid class

    This class gives data about the Glutamic Acid object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "E"

    def setState(self):
        """
        Set the name to use for the forcefield based on the current
        state.
        """
        if "GLH" in self.patches or self.name == "GLH":
            self.ffname = "GLH"
        Amino.setState(self)


class GLY(Amino):
    """
    Glycine class

    This class gives data about the Glycine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "G"


class HIS(Amino):
    """
    Histidine class

    This class gives data about the Histidine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "H"

    def setState(self):
        """
        Histidines are a special case due to the presence of
        several different forms.  This function sets all non-
        positive incarnations of HIS to neutral HIS by
        checking to see if optimization removed hacceptor or
        hdonor flags.  Otherwise HID is used as the default.
        """
        # if "HIP" not in self.patches and self.name not in ["HIP", "HSP"]:
        #    if self.getAtom("ND1").hdonor and not \
        #           self.getAtom("ND1").hacceptor:
        #        if self.hasAtom("HE2"): self.removeAtom("HE2")
        #    elif self.getAtom("NE2").hdonor and not \
        #             self.getAtom("NE2").hacceptor:
        #        if self.hasAtom("HD1"): self.removeAtom("HD1")
        #    elif self.getAtom("ND1").hacceptor and not \
        #             self.getAtom("ND1").hdonor:
        #        if self.hasAtom("HD1"): self.removeAtom("HD1")
        #    else: # Default to HID
        #        if self.hasAtom("HE2"): self.removeAtom("HE2")

        self.ffname = "HIP"
        Amino.setState(self)


class ILE(Amino):
    """
    Isoleucine class

    This class gives data about the Isoleucine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "I"


class LEU(Amino):
    """
    Leucine class

    This class gives data about the Leucine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "L"


class LYS(Amino):
    """
    Lysine class

    This class gives data about the Lysine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "K"

    def setState(self):
        """
        Determine if this is LYN or not
        """
        if "LYN" in self.patches or self.name == "LYN":
            self.ffname = "LYN"
        Amino.setState(self)


class MET(Amino):
    """
    Methionine class

    This class gives data about the Methionine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "M"


class PHE(Amino):
    """
    Phenylalanine class

    This class gives data about the Phenylalanine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "F"


class PRO(Amino):
    """
    Proline class

    This class gives data about the Proline object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "P"

    def setState(self):
        """
        Set the name to use for the forcefield based on the current
        state.  Uses N* and C* for termini.
        """
        if self.isNterm:
            self.ffname = "N%s" % self.ffname
        elif self.isCterm:
            if "NEUTRAL-CTERM" in self.patches:
                self.ffname = "NEUTRAL-C%s" % self.ffname
            else:
                self.ffname = "C%s" % self.ffname


class SER(Amino):
    """
    Serine class

    This class gives data about the Serine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "S"


class THR(Amino):
    """
    Threonine class

    This class gives data about the Threonine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "T"


class TRP(Amino):
    """
    Tryptophan class

    This class gives data about the Tryptophan object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "W"


class TYR(Amino):
    """
    Tyrosine class

    This class gives data about the Tyrosine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "Y"

    def setState(self):
        """
        See if the TYR is negative or not
        """
        if "TYM" in self.patches or self.name == "TYM":
            self.ffname = "TYM"
        Amino.setState(self)


class VAL(Amino):
    """
    Valine class

    This class gives data about the Valine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Amino.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "V"


class WAT(Residue):
    """
    Water class

    This class gives data about the Water object, and inherits
    off the base residue class.
    """

    water_residue_names = ["HOH", "WAT"]

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """

        sampleAtom = atoms[-1]

        self.atoms = []
        self.name = sampleAtom.resName
        self.chainID = sampleAtom.chainID
        self.resSeq = sampleAtom.resSeq
        self.iCode = sampleAtom.iCode

        self.fixed = 0
        self.ffname = "WAT"
        self.map = {}
        self.reference = ref

        # Create each atom

        for a in atoms:
            if a.name in ref.altnames:  # Rename atoms
                a.name = ref.altnames[a.name]

            atom = Atom(a, "HETATM", self)
            atomname = atom.get("name")
            if atomname not in self.map:
                self.addAtom(atom)
            else:  # Don't add duplicate atom with altLoc field
                oldatom = self.getAtom(atomname)
                oldatom.set("altLoc", "")

    def createAtom(self, atomname, newcoords):
        """
        Create a water atom.  Note the HETATM field.

        Parameters
            atomname: The name of the atom (string)
            newcoords:  The new coordinates of the atom (list)
        """
        oldatom = self.atoms[0]
        newatom = Atom(oldatom, "HETATM", self)
        newatom.set("x", newcoords[0])
        newatom.set("y", newcoords[1])
        newatom.set("z", newcoords[2])
        newatom.set("name", atomname)
        newatom.set("occupancy", 1.00)
        newatom.set("tempFactor", 0.00)
        newatom.added = 1
        self.addAtom(newatom)

    def addAtom(self, atom):
        """
        Override the existing addAtom - include the link to the
        reference object
        """
        self.atoms.append(atom)
        atomname = atom.get("name")
        self.map[atomname] = atom
        try:
            atom.reference = self.reference.map[atomname]
            for bond in atom.reference.bonds:
                if self.hasAtom(bond):
                    bondatom = self.map[bond]
                    if bondatom not in atom.bonds:
                        atom.bonds.append(bondatom)
                    if atom not in bondatom.bonds:
                        bondatom.bonds.append(atom)
        except KeyError:
            atom.reference = None


class LIG(Residue):
    """
    Generic ligand class

    This class gives data about the generic ligand object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """

        sampleAtom = atoms[-1]

        self.atoms = []
        self.name = sampleAtom.resName
        self.chainID = sampleAtom.chainID
        self.resSeq = sampleAtom.resSeq
        self.iCode = sampleAtom.iCode

        self.fixed = 0
        self.ffname = "WAT"
        self.map = {}
        self.reference = ref

        self.isNterm = 0
        self.isCterm = 0

        # Create each atom

        for a in atoms:
            if a.name in ref.altnames:  # Rename atoms
                a.name = ref.altnames[a.name]

            atom = Atom(a, "HETATM", self)
            atomname = atom.get("name")
            if atomname not in self.map:
                self.addAtom(atom)
            else:  # Don't add duplicate atom with altLoc field
                oldatom = self.getAtom(atomname)
                oldatom.set("altLoc", "")

    def createAtom(self, atomname, newcoords):
        """
        Create a water atom.  Note the HETATM field.

        Parameters
            atomname: The name of the atom (string)
            newcoords:  The new coordinates of the atom (list)
        """
        oldatom = self.atoms[0]
        newatom = Atom(oldatom, "HETATM", self)
        newatom.set("x", newcoords[0])
        newatom.set("y", newcoords[1])
        newatom.set("z", newcoords[2])
        newatom.set("name", atomname)
        newatom.set("occupancy", 1.00)
        newatom.set("tempFactor", 0.00)
        newatom.added = 1
        self.addAtom(newatom)

    def addAtom(self, atom):
        """
        Override the existing addAtom - include the link to the
        reference object
        """
        self.atoms.append(atom)
        atomname = atom.get("name")
        self.map[atomname] = atom
        try:
            atom.reference = self.reference.map[atomname]
            for bond in atom.reference.bonds:
                if self.hasAtom(bond):
                    bondatom = self.map[bond]
                    if bondatom not in atom.bonds:
                        atom.bonds.append(bondatom)
                    if atom not in bondatom.bonds:
                        bondatom.bonds.append(atom)
        except KeyError:
            atom.reference = None



"""
    Resinter extension

    Print interaction energy between each residue pair in the protein. 
"""

__date__ = "21 October 2011"
__authors__ = "Kyle Monson and Emile Hogan"

import extensions
from src.hydrogens import Optimize

# itertools FTW!
from itertools import product, permutations, izip, count
from src.hydrogens import hydrogenRoutines

# Here are the Ri -> [Ri0, Ri1] maps:
# ARG -> [{AR0}, {ARG}]
# ASP -> [{ASP}, {ASH}]
# GLU -> [{GLU}, {GLH}]
# CYS -> [{CYM}, {CYS}]
# LYS -> [{LYN}, {LYS}]
# TYR -> [{TYM}, {TYR}]
# CTERM -> [{CTERM}, {NEUTRAL-CTERM}]
# NTERM -> [{NEUTRAL-NTERM}, {NTERM}]
# HIS -> [{HSD, HSE}, {HIS}]
# HIP -> [{HID, HIE}, {HIP}]
# HSP -> [{HSD, HSE}, {HSP}]

_titrationSets = (
    (("AR0",), "ARG"),
    (("ASH",), "ASP"),
    (("CYX",), "CYS"),
    (("GLU",), "GLH"),
    (("HSD", "HSE"), "HSP"),
    (("HID", "HIE"), "HIP"),
    (("LYN",), "LYS"),
    (("TYM",), "TYR"),
    (("CTERM",), "NEUTRAL-CTERM"),
    (("NEUTRAL-NTERM",), "NTERM"),
)

_titrationSetsMap = {}

for tsSet in _titrationSets:
    for ts in tsSet[0]:
        _titrationSetsMap[ts] = tsSet

    _titrationSetsMap[tsSet[1]] = tsSet

# loose ends.
_titrationSetsMap["HIS"] = _titrationSetsMap["HSD"]
_titrationSetsMap["CYM"] = _titrationSetsMap["CYS"]


def usage():
    """
    Returns usage text for newresinter.
    """
    txt = "Print interaction energy between each residue pair in the protein to {output-path}.newresinter."
    return txt


def run_extension(routines, outroot, options):
    outname = outroot + ".newresinter"
    with open(outname, "w") as outfile:
        processor = ResInter(routines, outfile, options)
        processor.generate_all()
        processor.write_resinter_output()


class ResInter(object):
    def __init__(self, routines, outfile, options):
        self.pairEnergyResults = {}
        self.combinationCount = 0
        self.totalCombinations = 0
        self.options = options
        self.output = extensions.extOutputHelper(routines, outfile)
        self.routines = routines

    def save_interation_energy(self, first, second):
        energy = get_residue_interaction_energy(first, second)
        pairText = str(first) + " " + str(second)
        if pairText in self.pairEnergyResults:
            txt = "#%s re-tested!!! LOLWAT?\n" % pairText
            self.output.write(txt)
        else:
            self.pairEnergyResults[pairText] = energy

    def save_all_residue_interaction_energies(self):
        """
        Writes out the residue interaction energy for each possible
        residue pair in the protein.
        """
        residuepairs = permutations(self.routines.protein.getResidues(), 2)

        for pair in residuepairs:
            self.save_interation_energy(pair[0], pair[1])

    def save_one_with_all_interaction_energies(self, i):
        """
        Writes out the residue interaction energy for each possible
        residue pair in the protein.
        """
        residues = list(self.routines.protein.getResidues())
        target = residues[i]
        del residues[i]

        for residue in residues:
            self.save_interation_energy(target, residue)
            self.save_interation_energy(residue, target)

    def save_pair_interaction_energies(self, i, j):
        """
        Writes out the residue interaction energy for each possible
        residue pair in the protein.
        """
        residues = list(self.routines.protein.getResidues())

        self.save_interation_energy(residues[i], residues[j])
        self.save_interation_energy(residues[j], residues[i])

    def create_all_protonated(self):
        residueSet = get_residue_titration_set_protonated(
            self.routines.protein.getResidues()
        )
        self.process_residue_set(
            residueSet,
            clean=self.options.clean,
            neutraln=self.options.neutraln,
            neutralc=self.options.neutralc,
            ligand=self.options.ligand,
            assign_only=self.options.assign_only,
            chain=self.options.chain,
            debump=self.options.debump,
            opt=self.options.opt,
        )

        self.save_all_residue_interaction_energies()

    def create_all_single_unprotonated(self):
        combinations = residue_set_single_unprotonated_combinations(
            self.routines.protein.getResidues()
        )
        for residueSet, i in combinations:
            self.process_residue_set(
                residueSet,
                clean=self.options.clean,
                neutraln=self.options.neutraln,
                neutralc=self.options.neutralc,
                ligand=self.options.ligand,
                assign_only=self.options.assign_only,
                chain=self.options.chain,
                debump=self.options.debump,
                opt=self.options.opt,
            )

            self.save_one_with_all_interaction_energies(i)

    def create_all_pair_unprotonated(self):
        combinations = residue_set_pair_unprotonated_combinations(
            self.routines.protein.getResidues()
        )
        for residueSet, i, j in combinations:
            self.process_residue_set(
                residueSet,
                clean=self.options.clean,
                neutraln=self.options.neutraln,
                neutralc=self.options.neutralc,
                ligand=self.options.ligand,
                assign_only=self.options.assign_only,
                chain=self.options.chain,
                debump=self.options.debump,
                opt=self.options.opt,
            )

            self.save_pair_interaction_energies(i, j)

    def count_combinations(self):
        n = 0  # total iterable residues
        k = 0  # total iterable residues with two possible choices.

        allProtonated = get_residue_titration_set_protonated(
            self.routines.protein.getResidues()
        )

        for name in allProtonated:
            if name in _titrationSetsMap:
                n += 1

                if len(_titrationSetsMap[name][0]) == 2:
                    k += 1

        self.totalCombinations = (((n + k) ** 2) + (n - k) + 2) / 2

    def generate_all(self):
        """
        For every titration state combination of residue output the
        interaction energy for all possible residue pairs.
        """
        self.routines.write("Printing residue interaction energies...\n")

        self.count_combinations()

        # Phase 1: Everything protonated
        self.create_all_protonated()

        # Phase 2: Single unprotonated paired with everything else.
        self.create_all_single_unprotonated()

        # Phase 2: Pair unprotonated paired with each other.
        self.create_all_pair_unprotonated()

    def write_resinter_output(self):
        """
        Output the interaction energy between each possible residue pair.
        """
        for resultKey in sorted(self.pairEnergyResults.iterkeys()):
            self.output.write(
                resultKey + " " + str(self.pairEnergyResults[resultKey]) + "\n"
            )

        self.routines.write(
            str(self.combinationCount) + " residue combinations tried\n"
        )

    def process_residue_set(
        self,
        residueSet,
        clean=False,
        neutraln=False,
        neutralc=False,
        ligand=None,
        assign_only=False,
        chain=False,
        debump=True,
        opt=True,
    ):

        self.combinationCount += 1

        txt = "Running combination {0} of {1}\n".format(
            self.combinationCount, self.totalCombinations
        )
        self.routines.write(txt)

        self.routines.write(str(residueSet) + "\n")

        self.routines.removeHydrogens()

        for newResidueName, oldResidue, index in izip(
            residueSet, self.routines.protein.getResidues(), count()
        ):
            if newResidueName is None:
                continue

            chain = self.routines.protein.chainmap[oldResidue.chainID]
            chainIndex = chain.residues.index(oldResidue)
            residueAtoms = oldResidue.atoms

            # Create the replacement residue
            newResidue = self.routines.protein.createResidue(
                residueAtoms, newResidueName
            )

            # Make sure our names are cleaned up for output.
            newResidue.renameResidue(newResidueName)

            # Drop it in
            self.routines.protein.residues[index] = newResidue
            chain.residues[chainIndex] = newResidue

        # Run the meaty bits of PDB2PQR
        self.routines.setTermini(neutraln, neutralc)
        self.routines.updateBonds()

        if not clean and not assign_only:
            self.routines.updateSSbridges()

            if debump:
                self.routines.debumpProtein()

            self.routines.addHydrogens()

            hydRoutines = hydrogenRoutines(self.routines)

            if debump:
                self.routines.debumpProtein()

            if opt:
                hydRoutines.setOptimizeableHydrogens()
                hydRoutines.initializeFullOptimization()
                hydRoutines.optimizeHydrogens()
            else:
                hydRoutines.initializeWaterOptimization()
                hydRoutines.optimizeHydrogens()

            # Special for GLH/ASH, since both conformations were added
            hydRoutines.cleanup()


def get_residue_titration_set_protonated(residues):
    """
    Returns residue set when everything is protonated.
    """
    result = []
    for residue in residues:
        residueTest = _titrationSetsMap.get(residue.name)
        if residueTest:
            residueTest = residueTest[1]
        else:
            residueTest = residue.name
        result.append(residueTest)

    return result


def residue_set_single_unprotonated_combinations(residues):
    """
    Yields pair (residue set, residue index) for
    every "single unprotonated" combination.
    residue set - set for process_residue_set
    residue index - index of residue that was left unprotonated
    """
    protonatedNames = get_residue_titration_set_protonated(residues)

    for name, i in izip(protonatedNames, count()):
        if not name in _titrationSetsMap:
            continue

        tStateSet = _titrationSetsMap[name][0]

        for tState in tStateSet:
            result = list(protonatedNames)
            result[i] = tState
            yield result, i


def residue_set_pair_unprotonated_combinations(residues):
    """
    Yields pair (residue set, 1rst residue index, 2nd residue index) for
    every "single unprotonated" combination.
    residue set - set for process_residue_set
    1rst residue index - index of 1rst residue that was left unprotonated
    2nd residue index - index of 2nd residue that was left unprotonated
    """
    protonatedNames = get_residue_titration_set_protonated(residues)

    for i in xrange(0, len(protonatedNames)):
        firstName = protonatedNames[i]
        if not firstName in _titrationSetsMap:
            continue
        firstStateSet = _titrationSetsMap[firstName][0]
        for j in xrange(0, i):
            secondName = protonatedNames[j]
            if not secondName in _titrationSetsMap:
                continue

            secondStateSet = _titrationSetsMap[secondName][0]

            for firstState in firstStateSet:
                for secondState in secondStateSet:
                    result = list(protonatedNames)
                    result[i] = firstState
                    result[j] = secondState
                    yield result, i, j


def get_residue_interaction_energy(residue1, residue2):
    """
    Returns to total energy of every atom pair between the two residues.

    Uses Optimize.getPairEnergy and it's donor/accepter model
    to determine energy.

    residue1 - "donor" residue
    residue2 - "acceptor" residue

    THE RESULTS OF THIS FUNCTION ARE NOT SYMMETRIC. Swapping
    residue1 and residue2 will not always produce the same result.
    """
    energy = 0.0
    for pair in product(residue1.getAtoms(), residue2.getAtoms()):
        energy += Optimize.getPairEnergy(pair[0], pair[1])

    return energy
